import base64
from tonutils.client import TonapiClient
from tonutils.wallet.messages import TransferMessage
from typing import Optional, Any
from ..errors import WalletVersionError, WalletBalanceInsufficient

NANO = 1_000_000_000

class WalletClient:
  def __init__(self, tonapi_key, mnemonic, version: str = 'v5r1', is_testnet=False):
    self.api_key = tonapi_key
    self.is_testnet = is_testnet
    self._client = TonapiClient(api_key=self.api_key, is_testnet=self.is_testnet, rps=1)

    if version == 'v5r1':
      from tonutils.wallet import WalletV5R1
      self._wallet, _, _, _ = WalletV5R1.from_mnemonic(
        client=self._client,
        mnemonic=mnemonic
      )
    elif version == 'v4r2':
      from tonutils.wallet import WalletV4R2
      self._wallet, _, _, _ = WalletV4R2.from_mnemonic(
        client=self._client,
        mnemonic=mnemonic
      )
    else:
      raise WalletVersionError("Wrong wallet version specified. Use only v5r1 or v4r2")
  
  @property
  def address(self) -> str:
    """
      Returns user-friendly, non-bouncable wallet address
    """
    return self._wallet.address.to_str(True, True, False)
  
  async def get_balance(self) -> float:
    return await self._wallet.balance()
  
  def get_wallet_data(self) -> dict[str, Any]:
    """
      Returns object that is needed to create a purchase transaction
    """
    boc = self._wallet.state_init.serialize().to_boc()
    return {
      "address": self.address,
      "chain": "-239",
      "walletStateInit": base64.b64encode(boc).decode(),
      "publicKey": self._wallet.public_key.hex()
    }
  
  def _decode_payload(self, payload: str) -> str:
    from tonutils.utils import Slice, Cell
    
    payload = payload.strip()
    payload += "=" * ((4 - len(payload) % 4) % 4)
    
    boc_bytes = base64.b64decode(payload)
    cell = Cell.one_from_boc(boc_bytes)
    cs = cell.begin_parse()
    
    op = cs.load_uint(32)
    comment = cs.load_snake_string()
    return comment
  
  async def process_transaction(self, destination: str, amount: str, body: str) -> Optional[str]:
    if not self._wallet:
      raise ValueError("Wallet not connected.")
    
    amount = int(amount) / NANO
    wallet_balance = await self.get_balance()
    if wallet_balance < amount:
      raise WalletBalanceInsufficient("Wallet balance insufficient.")
    
    body = self._decode_payload(body)
    try:
      messages = [
        TransferMessage(
          destination=destination,
          amount=amount,
          body=body
        )
      ]
      tx_hash = await self._wallet.batch_transfer_messages(messages=messages)
      return tx_hash
    except Exception as error:
      raise Exception(error)


import requests

def http_request(url, method='GET', body=None, cookies=None):
  method = method.upper()
  
  headers = {
    "Accept": "application/json, text/javascript, */*; q=0.01",
    "Accept-Language": "en-US,en;q=0.9",
    "Connection": "keep-alive",
    "Cookie": cookies
  }
  
  response = None
  try:
    if method == "GET":
      response = requests.get(url, headers=headers)
    elif method == "POST":
      response = requests.post(url, data=body, headers=headers)
    else:
      raise ValueError('Only POST and GET methods!')
  except requests.exceptions.HTTPError as error:
    raise
  return response


from typing import Optional, Any, Tuple
from dataclasses import dataclass
from bs4 import BeautifulSoup
import re
import json
from requests.exceptions import HTTPError

from .methods import FragmentMethod
from .wallet import WalletClient
from .utils.http_request import http_request
from .errors import (
  FragmentInitError,
  FragmentAPIError,
  FragmentDataSetupError
)

@dataclass
class _Context:
  base_url: str
  cookies: str
  wallet: WalletClient
  ton_rate: Optional[float] = None

class FragmentClient:
  """Client for Fragment API.
  
    Automatically fetches session data during initialization.
  """
  
  def __init__(
    self,
    cookies: str,
    wallet: WalletClient
  ) -> None:
    if cookies is None or wallet is None:
      raise FragmentInitError('cookies and wallet cannot be None')
    
    self._ctx = _Context(base_url='https://fragment.com', cookies=cookies, wallet=wallet)
    self._setup_client_data()
  
  @property
  def stars(self) -> "FragmentStarsAPI":
    from .stars import FragmentStarsAPI
    return FragmentStarsAPI(self)
  
  @property
  def premium(self) -> "FragmentPremiumAPI":
    from .premium import FragmentPremiumAPI
    return FragmentPremiumAPI(self)
  
  @property
  def ads(self) -> "FragmentAdsAPI":
    from .ads import FragmentAdsAPI
    return FragmentAdsAPI(self)
  
  def _setup_client_data(self) -> None:
    try:
      response = http_request(self._ctx.base_url, method='GET', cookies=self._ctx.cookies)
      html_content = response.text
      
      soup = BeautifulSoup(html_content, 'html.parser')
      script_tags = soup.find_all('script')
      
      if match := re.search(r'ajInit\(({.+?})\);', html_content):
        json_str = match.group(1)
        json_obj = json.loads(json_str)
        api_url = json_obj.get('apiUrl')
        self._ctx.base_url = self._ctx.base_url + api_url
        
        if state := json_obj.get('state'):
          self._ctx.ton_rate = state.get('tonRate')
    except HTTPError as error:
      raise FragmentDataSetupError("Error setting up client data.")
  
  def send_fragment_request(self, method: FragmentMethod, body) -> Tuple[Optional[dict], Optional[str]]:
    try:
      response = http_request(
        self._ctx.base_url,
        method='POST',
        body={**body, 'method': method.value},
        cookies=self._ctx.cookies
      )
      data = response.json()
      
      if error := data.get('error'):
        return None, error
      
      return data, None
    except HTTPError as error:
      raise FragmentAPIError("Failed to complete Fragment API request.", str(error))
  
  @property
  def base_url(self) -> str:
    return self._ctx.base_url

  @property
  def ton_rate(self) -> Optional[float]:
    return self._ctx.ton_rate


class FragmentClientError(Exception):
  """Base class for Client errors"""

class FragmentDataSetupError(FragmentClientError):
  """Error raised when there's a problem with setting up Fragment client data:
  (api_hash, ton_rate, etc).
  """

class FragmentAPIError(FragmentClientError):
  """Error raised when Fragment does not send the expected response."""
  def __init__(self, message, response):
    msg = f"{message}\nThe server responded with: {response}"
    self.response = response
    super(FragmentAPIError, self).__init__(msg)

class FragmentInitError(FragmentClientError):
  """Raised when a required parameter is missing or invalid."""

class StarsOperationError(FragmentClientError):
  """Errors in Stars operations"""

class PremiumOperationError(FragmentClientError):
  """Errors in Telegram Premium operations"""

class AdsOperationError(FragmentClientError):
  """Errors in Telegram Ads operations"""

class WalletError(FragmentClientError):
  """Raised when there's a problem with wallet."""

class WalletVersionError(WalletError):
  """Raised when wrong wallet version specified"""

class WalletBalanceInsufficient(WalletError):
  """Raised when wallet balance insufficient"""


from typing import Tuple, Optional, Any
from .utils.extract_photo_src import extract_photo_src
from .methods import FragmentMethod
from .errors import StarsOperationError, FragmentAPIError, WalletBalanceInsufficient
from .models import (
  ApiResult,
  SearchRecipientResponseData,
  InitPurchaseResponseData,
  Transaction,
  TransactionMessage,
)

class FragmentStarsAPI:
  def __init__(self, client: "FragmentClient") -> None:
    self._client = client
    self._wallet = self._client._ctx.wallet
  
  def init_buy_stars_request(self, recipient_id, quantity) -> ApiResult[InitPurchaseResponseData]:
    try:
      data, error = self._client.send_fragment_request(
        method=FragmentMethod.INIT_BUY_STARS,
        body={ 'recipient': recipient_id, 'quantity': quantity }
      )
      
      if error:
        return ApiResult.fail(f'Failed to initialize Stars purchase: {error}')
      
      return ApiResult.ok(
        data=InitPurchaseResponseData(
          req_id=data.get("req_id"),
          amount=data.get("amount"),
          to_bot=bool(data.get("to_bot"))
        )
      )
    except FragmentAPIError as error:
      raise StarsOperationError(f"Failed to initialize Stars purchase: {error}")
  
  def get_buy_stars_link(self, account, req_id) -> ApiResult[Transaction]:
    try:
      data, error = self._client.send_fragment_request(
        method=FragmentMethod.GET_BUY_STARS_LINK,
        body={
          'account': account,
          'transaction': '1',
          'id': req_id,
          'show_sender': '0'
        }
      )
      
      if error:
        return ApiResult.fail(f'Failed to obtain Stars purchase link: {error}')
      
      transaction = data.get('transaction')
      transaction_messages = transaction.get('messages')
      messages = [
        TransactionMessage(
          address=item.get('address'),
          amount=item.get('amount'),
          payload=item.get('payload')
        )
        for item in transaction_messages
      ]
      
      return ApiResult.ok(
        data=Transaction(
          validUntil=transaction.get('validUntil'),
          messages=messages
        )
      )
    except FragmentAPIError as error:
      raise StarsOperationError(f"Failed to obtain Stars purchase link: {error}")
  
  def search_stars_recipient(self, recipient_username) -> ApiResult[SearchRecipientResponseData]:
    try:
      data, error = self._client.send_fragment_request(
        method=FragmentMethod.SEARCH_STARS_RECIPIENT,
        body={
          'query': recipient_username,
          'quantity': '',
        }
      )
      
      if error:
        return ApiResult.fail(f"Failed to obtain Stars recipient: {error}")
      
      if found := data.get('found'):
        photo = found.get('photo')
        photo_src = extract_photo_src(html_photo=photo)
      
      return ApiResult.ok(
        data=SearchRecipientResponseData(
          recipient_id=found.get('recipient'),
          name=found.get('name'),
          photo_url=photo_src
        )
      )
    except FragmentAPIError as error:
      raise StarsOperationError(f"Failed to obtain Stars recipient: {error}")
  
  def create_stars_transaction(self, recipient_id, quantity, wallet_account) -> ApiResult[Transaction]:
    try:
      init_data = self.init_buy_stars_request(
        recipient_id=recipient_id,
        quantity=quantity
      )
      
      if init_data.error:
        return ApiResult.fail(init_data.error)
      
      buy_transaction = self.get_buy_stars_link(
        account=wallet_account,
        req_id=init_data.data.req_id
      )
      
      if buy_transaction.error:
        return ApiResult.fail(buy_transaction.error)
      
      return ApiResult.ok(data=buy_transaction.data)
    except StarsOperationError:
      raise
  
  async def buy_stars(self, username: str, quantity: int) -> ApiResult[str]:
    try:
      recipient = self.search_stars_recipient(recipient_username=username)
      if recipient.error:
        return ApiResult.fail(recipient.error)
      wallet_account = self._wallet.get_wallet_data()
      transaction = self.create_stars_transaction(recipient_id=recipient.data.recipient_id, quantity=quantity, wallet_account=wallet_account)
      
      if transaction.error:
        return ApiResult.fail(transaction.error)

      transaction_msg = transaction.data.messages[0]
      tx_hash = await self._wallet.process_transaction(
        destination=transaction_msg.address,
        amount=transaction_msg.amount,
        body=transaction_msg.payload
      )
      
      return ApiResult.ok(data=tx_hash)
    except WalletBalanceInsufficient as error:
      return ApiResult.fail(error=error)
    except StarsOperationError:
      raise


from dataclasses import dataclass, asdict
from typing import Any, Generic, Optional, TypeVar, List

T = TypeVar("T")

class BaseResponseModel:
  def to_dict(self) -> dict[str, Any]:
    return asdict(self)

@dataclass(slots=True)
class ApiResult(Generic[T]):
  success: bool
  data: Optional[T] = None
  error: Optional[str] = None
  
  @classmethod
  def ok(cls, data: T) -> "ApiResult[T]":
    return cls(success=True, data=data)

  @classmethod
  def fail(cls, error: str) -> "ApiResult[T]":
    return cls(success=False, error=error)

  def __bool__(self) -> bool:
    return self.success

@dataclass
class SearchRecipientResponseData(BaseResponseModel):
  recipient_id: str
  name: str
  photo_url: str | None

@dataclass
class InitPurchaseResponseData(BaseResponseModel):
  req_id: str
  amount: str
  to_bot: Optional[bool] = None

@dataclass
class TransactionMessage(BaseResponseModel):
  address: str
  amount: str
  payload: str

@dataclass
class Transaction(BaseResponseModel):
  validUntil: int
  messages: List[TransactionMessage]


from enum import Enum

class FragmentMethod(Enum):
  INIT_BUY_STARS = 'initBuyStarsRequest'
  GET_BUY_STARS_LINK = 'getBuyStarsLink'
  SEARCH_STARS_RECIPIENT = 'searchStarsRecipient'
  
  INIT_GIFT_PREMIUM = 'initGiftPremiumRequest'
  GET_GIFT_PREMIUM_LINK = 'getGiftPremiumLink'
  SEARCH_PREMIUM_RECIPIENT = 'searchPremiumGiftRecipient'
  
  INIT_ADS_TOPUP = 'initAdsTopupRequest'
  GET_ADS_TOPUP_LINK = 'getAdsTopupLink'
  SEARCH_ADS_RECIPIENT = 'searchAdsTopupRecipient'